#include <fstream>
#include <iostream>
#include <vector>
#include <sstream>
#include <map>
#include <stdlib.h>
#include <algorithm> //min_element, max_element, nth_element
#include <boost/math/special_functions/gamma.hpp>
using namespace std;

const int NO_ERROR = 0;
int read_resample_files( const std::vector< string > & filename_vec, 
                         const std::vector< bool > & doBurden_vec,
                         double alpha,
                         double & threshold )
{
    if ( ( alpha < 0.0 ) || ( alpha > 1.0 ) )
    {
        cout << "read_resample_files error : alpha should be " <<
            "between 0.0 and 1.0.  alpha == " << alpha << endl;
        return -1;
    }    
    if ( filename_vec.size() != doBurden_vec.size() )
    {
        cout << "read_resample_files error : filename_vec and doBurden_vec"
             << " have unequal lengths.  filename_vec.size() == "
             << filename_vec.size() << "doBurden_vec.size() == " 
             << doBurden_vec.size();
        return -2;
    }
    if ( doBurden_vec.size() == 0 )
    {
        cout << "read_resample_files error : filename_vec and "
             << "doBurden_vec have length 0" << endl;
        return -3;
    }
    
    std::vector< std::vector< double > > pvals_for_each_repl;
    size_t old_num_lines = 0;
    for ( size_t file_index = 0; file_index < filename_vec.size(); 
            ++file_index )
    {
        const string in_filename = filename_vec[ file_index ];
        ifstream in_file( in_filename.c_str() );  
        if (!in_file.is_open())
        {
            cout << "read_resample_files : ";
            cout << "error : could not open " << in_filename <<endl;        
            return -4;
        }

        const bool doBurden = doBurden_vec[ file_index ];
        string line;
        size_t line_num = 0;
        while (getline(in_file,line))
        {
            string buff;
            stringstream ss(line);
            vector <double> line_elements;
            while (ss >> buff)
            {
                double element = atof(buff.c_str() );
                line_elements.push_back(element);
            }
            double pval_for_this_line;
            if ( doBurden )
            {
                auto val_for_this_line = std::max_element(
                    std::begin( line_elements ),
                    std::end(  line_elements )
                );
                //*val_for_this_line is a Burden Q value
                //std normal
                pval_for_this_line = 
                        boost::math::gamma_q( 
                            0.5,( *val_for_this_line / 2.0) );
            }
            else 
            {   //SKAT
                auto val_for_this_line =
                    std::min_element(
                    std::begin( line_elements ),
                    std::end(  line_elements )
                );
                pval_for_this_line = *val_for_this_line;
            }
            if ( file_index == 0 )
            {
                std::vector< double > my_vec;
                my_vec.push_back( pval_for_this_line );
                pvals_for_each_repl.push_back( my_vec );
            }
            else
            {
               std::vector<double> & my_vec = 
                    pvals_for_each_repl[ line_num ]; 
                my_vec.push_back( pval_for_this_line );
            }
            line_num++;
        } //while (getline(in_file,line)
        in_file.close();
        if ( file_index != 0 )
        {
            if ( line_num != old_num_lines )
            {
                cout << "read_resample_files : " <<
                    "error : number of replicates should be the same " <<
                    "for all files" << endl;
                return -5;
            }            
        }
        old_num_lines = line_num;
    } //for file_index
    //next get the minimum value in each of the constituent
    //vectors of pvals_for_each_repl
    vector< double > min_pval_for_each_repl_vec;
    for ( const auto & my_vec : pvals_for_each_repl )
    {
      auto my_min =
                    std::min_element(
                    std::begin( my_vec ),
                    std::end(  my_vec  )
                    );
      min_pval_for_each_repl_vec.push_back( *my_min );  
    }

    size_t desired_index;
    desired_index = ( alpha * min_pval_for_each_repl_vec.size() ) - 1;

    std::nth_element( min_pval_for_each_repl_vec.begin(),
                      min_pval_for_each_repl_vec.begin() + desired_index,
                      min_pval_for_each_repl_vec.end() );
    threshold = min_pval_for_each_repl_vec[ desired_index ];

    return NO_ERROR;
}

int read_listfile( 
    const string & listfilename, std::vector<string> & filename_vec,
    std::vector<bool> & doBurden_vec //Burden if bool is true 
                                    //otherwise SKAT 
    )
{
    ifstream in_file( listfilename );
    if (!in_file.is_open())
    {           
        cout << "read_listfile : ";
        cout << "error : could not open " << listfilename << endl;
        return -1;
    }
        cout << "Reading input from " << listfilename <<endl;              
        string line;    
        doBurden_vec.clear();
        filename_vec.clear();
        while (getline(in_file,line))
        {
            string buff;
            stringstream ss(line);
            vector <string> line_elements;
            while (ss >> buff)
            {
                line_elements.push_back(buff);
            }
            const size_t num_elements = line_elements.size();
            if ( num_elements == 0 )
            {
                continue;
            }
            if ( ( num_elements != 1  ) && ( num_elements != 2 ) )
            {
                in_file.close();
                cout << "read_listfile : line in input file should have 0, "
                    << " 1, or 2 columns but it has " << line_elements.size() 
                        << 
                    endl;
                return -2;
            }
           bool doBurden;
           //cout << "num_elements == " << num_elements << endl;
           if ( num_elements == 1 )
           {
             //assume Burden or single variant
             doBurden = true;
             cout << "No simulation type is given in " << listfilename 
                  << " for " << line_elements[0] << endl;
             cout << "Assuming " << line_elements[0]  << " is from either Burden or "
                  << "single-variant simulation."
                   << endl; 
           }
           else 
           {
               if ( ( line_elements[ 1 ] != "s" ) && 
                    ( line_elements[ 1 ] != "S" ) &&
                    ( line_elements[ 1 ] != "B" ) &&
                    ( line_elements[ 1 ] != "b" ) )
               {
                   in_file.close();
                   cout << "Error: the simulation type should be S or s for SKAT or "
                        << "B or b for either Burden or Single-Variant." << endl
                        << "Found " << line_elements[1] << " in " << listfilename
                        << " for " << line_elements[0] << endl;
                   return -4;
               }

               if ( ( line_elements[ 1 ] == "s" )  || ( line_elements[ 1 ] == "S" ) )
               {
                  doBurden = false;
                  cout << "Data in " << line_elements[0] << " is from SKAT simulation."
                       << endl;
               }
               else
                if ( ( line_elements[ 1 ] == "b" )  || ( line_elements[ 1 ] == "B" ) )
               {
                 doBurden = true;
                 cout << "Data in " << line_elements[0] << " is from either Burden or "
                      << "single-variant analysis." << endl;
               }
              else
               {
                 in_file.close();
                 cout << "Error: invalid value " << line_elements[ 1 ] 
                      << "given for simulation-type for file "
                      << line_elements[ 0 ] << endl;
                 return -3;
               }
           }
            //single variant test should be 
            //handled similarly to a Burden test in this code

            doBurden_vec.push_back( doBurden );
            filename_vec.push_back( line_elements[ 0 ] );
        } //while

        in_file.close();
    return NO_ERROR;
}



int main( int argc, char * argv[] )
{
    //cout << "argc == " << argc << endl;
    if ( argc != 2 )
    {
        cout << "Usage: " <<
         argv[0] << " listfilename" << endl;
        return -1;
    }

    stringstream convert_listfilename(argv[1]);
    string listfilename;
    if (!(convert_listfilename >> listfilename))
    {              
        cout << "Could not convert command line ";
        cout << "argument to string type for listfilename" << endl;
        return -2;
    }

    int rv;
    std::vector< string > filename_vec;
    std::vector< bool > doBurden_vec;
    rv = read_listfile( listfilename, filename_vec, doBurden_vec );
    if ( rv != NO_ERROR )
    {
        cout << "error: read_listfile returned == " << rv << endl;
        return EXIT_FAILURE;
    }
    cout << "Resample files" << endl;
    for (size_t i = 0; i < filename_vec.size(); ++i)
    {
        cout << filename_vec[ i ] << endl;
    }
    double threshold = 42;
    const double SIG_LEVEL = 0.05;

    rv = read_resample_files( filename_vec,
                              doBurden_vec,
                              0.05,
                              threshold );
    if ( rv != NO_ERROR)
    {
        cout << "error : read_resample_files returned " << rv << endl;
        return EXIT_FAILURE;
    }
    cout << "Significance level : " << SIG_LEVEL << endl;
    cout << "Threshold : " << threshold << endl;
    return EXIT_SUCCESS;
}



